# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
R"""This subpackage provides various tools for working with the geometric
representation of quaternions. A particular focus is computing the distance
between quaternions. These distance computations can be complicated,
particularly good metrics for distance on the Riemannian manifold representing
quaternions do not necessarily coincide with good metrics for similarities
between rotations. An overview of distance measurements can be found in
`this paper <https://link.springer.com/article/10.1007/s10851-009-0161-2>`_.
"""
from __future__ import division, print_function, absolute_import

import numpy as np

from ..functions import norm, exp, multiply, inverse, log, _validate_unit

__all__ = ['distance',
           'sym_distance',
           'riemann_exp_map',
           'riemann_log_map',
           'intrinsic_distance',
           'sym_intrinsic_distance',
           'angle']


def distance(p, q):
    R"""Determine the distance between quaternions p and q.

    This is the most basic distance that can be defined on
    the space of quaternions; it is the metric induced by
    the norm on this vector space
    :math:`\rho(p, q) = \lvert\lvert p - q \rvert\rvert`.

    When applied to unit quaternions, this function produces
    values in the range :math:`[0, 2]`.

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.

    Returns:
        Array of shape (...) containing the element-wise distances between the
        two sets of quaternions.

    Example::

        rowan.geometry.distance([1, 0, 0, 0], [1, 0, 0, 0])
    """
    return norm(np.asarray(p) - np.asarray(q))


def sym_distance(p, q):
    R"""Determine the distance between quaternions p and q.

    This is a symmetrized version of :py:func:`distance` that
    accounts for the fact that :math:`p` and :math:`-p` represent
    identical rotations. This makes it a useful measure of rotation
    similarity.

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.

    When applied to unit quaternions, this function produces
    values in the range :math:`[0, \sqrt{2}]`.

    Returns:
        Array of shape (...) containing the element-wise symmetrized distances
        between the two sets of quaternions.

    Example::

        rowan.geometry.sym_distance([1, 0, 0, 0], [-1, 0, 0, 0])
    """
    p = np.asarray(p)
    q = np.asarray(q)
    return np.minimum(norm(p - q), norm(p + q))


def riemann_exp_map(p, v):
    R"""Compute the exponential map on the Riemannian manifold
    :math:`\mathbb{H}^*` of nonzero quaterions.

    The nonzero quaternions form a Lie algebra :math:`\mathbb{H}^*` that
    is also a Riemannian manifold. In general, given a point :math:`p` on a
    Riemannian manifold :math:`\mathcal{M}` and an element of the tangent
    space at :math:`p`, :math:`v \in T_p\mathcal{M}`, the Riemannian exponential
    map is defined by the geodesic starting at :math:`p` and tracing out
    an arc of length :math:`v` in the direction of :math:`v`. This function
    computes the endpoint of that path (which is itself a quaternion).

    Explicitly, we define the exponential map as

    .. math::
        \begin{equation}
            \textrm{Exp}_p(v) = p\exp(v)
        \end{equation}

    Args:
        p ((...,4) np.array): Points on the manifold of quaternions.
        v ((...,4) np.array): Tangent vectors to traverse.

    Returns:
        Array of shape (..., 4) containing the endpoints of the geodesic
        starting from :math:`p` and traveling a distance :math:`\lvert\lvert
        v\rvert\rvert` in the direction of :math:`v`.

    Example::

        rowan.geometry.riemann_exp_map([1, 0, 0, 0], [-1, 0, 0, 0])
    """
    return multiply(p, exp(v))


def riemann_log_map(p, q):
    R"""Compute the log map on the Riemannian manifold :math:`\mathbb{H}^*` of
    nonzero quaterions.

    This function inverts :py:func:`riemann_exp_map`. See that function for more
    details. In brief, given two quaternions p and q, this method returns a
    third quaternion parameterizing the geodesic passing from p to q. It is
    therefore an important measure of the distance between the two input
    quaternions.

    Args:
        p ((...,4) np.array): Starting points (quaternions).
        q ((...,4) np.array): Endpoints (quaternions).

    Returns:
        Array of shape (..., 4) containing quaternions pointing from p to q with
        magnitudes equal to the length of the geodesics joining these
        quaternions.

    Example::

        rowan.geometry.riemann_log_map([1, 0, 0, 0], [-1, 0, 0, 0])
    """
    return log(multiply(inverse(q), p))


def intrinsic_distance(p, q):
    R"""Compute the intrinsic distance between quaternions on the manifold of
    quaternions.

    The quaternion distance is determined as the length of the quaternion
    joining the two quaternions (see :py:func:`riemann_log_map`). Rather
    than computing this directly, however, as shown in [Huynh09]_ we can
    compute this distance using the following equivalence:

    .. math::
        \begin{equation}
            \lvert\lvert \log(p q^{-1}) \rvert\rvert =
            2\cos(\lvert\langle p, q \rangle\rvert)
        \end{equation}

    When applied to unit quaternions, this function produces
    values in the range :math:`[0, \pi]`.

    .. [Huynh09] Huynh DQ (2009) Metrics for 3D rotations: comparison and
        analysis. J Math Imaging Vis 35(2):155-164

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.

    Returns:
        Array of shape (...) containing the element-wise intrinsic distances
        between the two sets of quaternions.

    Example::

        rowan.geometry.intrinsic_distance([1, 0, 0, 0], [-1, 0, 0, 0])
    """
    # TODO: Consider implementing the optimization
#    if not np.allclose(2*np.arccos(np.linalg.norm(np.inner(p, q))),
#            norm(riemann_log_map(p, q))):
#        raise ValueError("Huh?")
#    return 2*np.arccos(np.linalg.norm(np.inner(p, q))),
    return norm(riemann_log_map(p, q))


def sym_intrinsic_distance(p, q):
    R"""Compute the intrinsic distance between quaternions on the manifold of
    quaternions.

    This is a symmetrized version of :py:func:`intrinsic_distance` that
    accounts for the double cover :math:`SU(2)\rightarrow SO(3)`, making it a
    more useful metric for rotation similarity.

    When applied to unit quaternions, this function produces
    values in the range :math:`[0, \frac{\pi}{2}]`.

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.

    Returns:
        Array of shape (...) containing the element-wise symmetrized intrinsic
        distances between the two sets of quaternions.

    Example::

        rowan.geometry.sym_intrinsic_distance([1, 0, 0, 0], [-1, 0, 0, 0])
    """
    p = np.asarray(p)
    q = np.asarray(q)
    return np.where(norm(p - q) < norm(p + q),
                    norm(riemann_log_map(p, q)),
                    norm(riemann_log_map(p, -q))
                    )


def angle(p):
    R"""Compute the angle of rotation of a quaternion.

    Note that this is identical to
    ``intrinsic_distance(p, np.array([1, 0, 0, 0]))``.

    Args:
        p ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing the element-wise angles traced out by
        these rotations.

    Example::

        rowan.geometry.angle([1, 0, 0, 0])
    """

    # TODO: Make sure all the quaternions are rotations
    # where they need to be.

    _validate_unit(p)
    return 2*norm(log(p))



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""
Various functions for generating random sets of rotation quaternions. Note
that if you simply want random quaternions not restricted to :math:`SO(3)` you
can just generate these directly using ``np.random.rand(... 4)``. This
subpackage is entirely focused on generating rotation quaternions.
"""

# TODO: Try implementing randomness using this paper or one of the ones it
# cites:
#    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2896220/
#
#    other alternatives:
#    http://home.lu.lv/~sd20008/papers/essays/Random%20unitary%20[paper].pdf
from __future__ import division, print_function, absolute_import

import numpy as np

__all__ = ['rand',
           'random_sample'
           ]


def rand(*args):
    R"""Generate random rotations that are uniformly distributed on a unit sphere.

    This is a convenience function *a la* ``np.random.rand``. If you want a
    function that takes a tuple as input, use :py:func:`random_sample` instead.

    Args:
        shape (tuple): The shape of the array to generate.

    Return:
        Random quaternions of the shape provided with an additional axis of
        length 4.

    Example::

        q_rand = rowan.random.rand(3, 3, 2)
    """
    if len(args) == 0:
        return random_sample()
    else:
        return random_sample(args)


def random_sample(size=None):
    R"""Generate random rotations uniformly

    In general, sampling from the space of all quaternions will not generate
    uniform rotations. What we want is a distribution that accounts for the
    density of rotations, *i.e.*, a distribution that is uniform with respect
    to the appropriate measure. The algorithm used here is detailed in
    [Shoe92]_.

    .. [Shoe92] Shoemake, K.: Uniform random rotations. In: D. Kirk, editor,
        Graphics Gems III, pages 124-132. Academic, New York, 1992.

    Args:
        size (tuple): The shape of the array to generate.

    Return:
        Random quaternions of the shape provided with an additional axis of
        length 4.

    Example::

        q_rand = rowan.random.random_sample((3, 3, 2))
    """
    if size is None:
        size = (3,)
    else:
        size += (3,)

    u = np.random.random_sample(size)

    theta1 = 2*np.pi*u[..., 1]
    s1 = np.sin(theta1)
    c1 = np.cos(theta1)
    theta2 = 2*np.pi*u[..., 2]
    s2 = np.sin(theta2)
    c2 = np.cos(theta2)

    r1 = np.sqrt(1-u[..., 0])
    r2 = np.sqrt(u[..., 0])

    quats = np.stack((s1*r1, c1*r1, s2*r2, c2*r2), axis=-1)
    return quats.squeeze()



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
R"""
The core :py:mod:`rowan` package contains functions for operating on
quaternions. The core package is focused on robust implementations of key
functions like multiplication, exponentiation, norms, and others. Simple
functionality such as addition is inherited directly from NumPy due to
the representation of quaternions as NumPy arrays. Many core NumPy functions
implemented for normal arrays are reimplemented to work on quaternions (
such as :py:func:`allclose` and :py:func:`isfinite`). Additionally, `NumPy
broadcasting
<https://docs.scipy.org/doc/numpy-1.14.0/user/basics.broadcasting.html#>`_
is enabled throughout rowan unless otherwise specified. This means that
any function of 2 (or more) quaternions can take arrays of shapes that do
not match and return results according to NumPy's broadcasting rules.
"""

from __future__ import division, print_function, absolute_import

from . import calculus
from . import geometry
from . import interpolate
from . import mapping
from . import random

from .functions import (allclose, conjugate, divide, exp, expb, exp10, equal,
                        from_axis_angle, from_euler, from_matrix,
                        from_mirror_plane, inverse, isclose, isinf, isfinite,
                        isnan, is_unit, log, logb, log10, multiply, norm,
                        normalize, not_equal, power, reflect, rotate,
                        to_axis_angle, to_euler, to_matrix,
                        vector_vector_rotation)

# Get the version
__version__ = '1.1.7'

__all__ = ['calculus', 'geometry', 'interpolate', 'mapping', 'random',
           'allclose', 'conjugate', 'divide', 'exp', 'expb', 'exp10', 'equal',
           'from_axis_angle', 'from_euler', 'from_matrix', 'from_mirror_plane',
           'inverse', 'isclose', 'isinf', 'isfinite', 'isnan', 'is_unit', 'log',
           'logb', 'log10', 'multiply', 'norm', 'normalize', 'not_equal',
           'power', 'reflect', 'rotate', 'to_axis_angle', 'to_euler',
           'to_matrix', 'vector_vector_rotation']



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
R"""
The general space of problems that this subpackage addresses is a small subset
of the broader space of `point set registration
<https://en.wikipedia.org/wiki/Point_set_registration>`_, which attempts to
optimally align two sets of points. In general, this mapping can be nonlinear.
The restriction of this superposition to linear transformations composed of
translation, rotation, and scaling is the study of Procrustes superposition,
the first step in the field of `Procrustes analysis
<https://en.wikipedia.org/wiki/Procrustes_analysis#Shape_comparison>`_, which
performs the superposition in order to compare two (or more) shapes.

If points in the two sets have a known correspondence, the problem is much
simpler. Various precise formulations exist that admit analytical formulations,
such as the `orthogonal Procrustes problem
<https://en.wikipedia.org/wiki/Orthogonal_Procrustes_problem>`_ searching for an
orthogonal transformation

.. math::
    \begin{equation}
        R = \textrm{argmin}_\Omega \lvert\lvert\Omega A - B\rvert\rvert_F,\,\,
        \Omega^T\Omega = \mathbb{1}
    \end{equation}

or, if a pure rotation is desired, Wahba's problem

.. math::
    \begin{equation}
        \min_{\boldsymbol{R} \in SO(3)} \frac{1}{2} \sum_{k=1}^N a_k \lvert
        \lvert \boldsymbol{w}_k - \boldsymbol{R} \boldsymbol{v}_k \rvert\rvert^2
    \end{equation}

Numerous algorithms to solve this problem exist, particularly in the field of
aerospace engineering and robotics where this problem must be solved on embedded
systems with limited processing. Since that constraint does not apply here, this
package simply implements some of the most stable known methods irrespective of
cost. In particular, this package contains the Kabsch algorithm, which solves
Wahba's problem using an SVD in the vein of `Peter Schonemann's original
solution <https://link.springer.com/article/10.1007/BF02289451>`_ to
the orthogonal Procrustes problem. Additionally this package contains the
`Davenport q method <https://ntrs.nasa.gov/search.jsp?R=19670009376>`_, which
works directly with quaternions. The most popular algorithms for Wahba's problem
are variants of the q method that are faster at the cost of some stability; we
omit these here.

In addition, :py:mod:`rowan.mapping` also includes some functionality for
more general point set registration. If a point cloud has a set of known
symmetries, these can be tested explicitly by :py:mod:`rowan.mapping` to
find the smallest rotation required for optimal mapping. If no such
correspondence is knowna at all, then the iterative closest point algorithm can
be used to approximate the mapping.
"""
from __future__ import division, print_function, absolute_import

import numpy as np

from ..functions import from_matrix, rotate, to_matrix
from ..geometry import angle

__all__ = ['kabsch', 'davenport', 'procrustes', 'icp']


def kabsch(X, Y, require_rotation=True):
    R"""Find the optimal rotation and translation to map between two sets of
    points.

    This function implements the
    `Kabsch algorithm <https://en.wikipedia.org/wiki/Kabsch_algorithm>`_, which
    minimizes the RMSD between two sets of points. One benefit of this approach
    is that the SVD works in dimensions > 3.

    Args:
        X ((N, m) np.array): First set of N points.
        Y ((N, m) np.array): Second set of N points.
        require_rotation (bool): If false, the returned quaternion.

    Returns:
        A tuple (R, t) where R is the (m x m) rotation matrix to rotate the
        points and t is the translation.

    Example::

        import numpy as np

        # Create some random points, then make a random transformation of
        # these points
        points = np.random.rand(10, 3)
        rotation = rowan.random.rand(1)
        translation = np.random.rand(1, 3)
        transformed_points = rowan.rotate(rotation, points) + translation

        # Recover the rotation and check
        R, t = rowan.mapping.kabsch(points, transformed_points)
        q = rowan.from_matrix(R)

        assert np.logical_or(
            np.allclose(rotation, q), np.allclose(rotation, -q))
        assert np.allclose(translation, t)
    """
    X = np.atleast_2d(X)
    Y = np.atleast_2d(Y)
    if X.shape != Y.shape:
        raise ValueError("Input arrays must be the same shape")
    elif len(X.shape) != 2:
        raise ValueError("Input arrays must be (Nxm) arrays")
    X = np.asarray(X)
    Y = np.asarray(Y)

    # The algorithm depends on removing the centroid of the points.
    centroid_X = np.mean(X, axis=0)
    centroid_Y = np.mean(Y, axis=0)
    X_c = X - centroid_X
    Y_c = Y - centroid_Y

    H = X_c.T.dot(Y_c)

    U, S, Vt = np.linalg.svd(H)

    R = Vt.T.dot(U.T)

    # Remove reflections by negating final eigenvalue
    if np.linalg.det(R) < 0:
        Vt[2, :] *= -1
        R = Vt.T.dot(U.T)

    t = -R.dot(centroid_X.T) + centroid_Y.T

    return R, t


def horn(X, Y):
    R"""Find the optimal rotation and translation to map between two sets of
    points.

    This function implements the quaternion-based method of `Horn
    <https://www.osapublishing.org/josaa/abstract.cfm?id=2711>`_. For a simpler
    explanation, see `here
    <https://cnx.org/contents/HV-RsdwL@23/Molecular-Distance-Measures>`_.

    Args:
        X ((N, 3) np.array): First set of N points.
        Y ((N, 3) np.array): Second set of N points.

    Returns:
        A tuple (q, t) where q is the quaternion to rotate the points and t
        is the translation.

    Example::

        import numpy as np

        # Create some random points, then make a random transformation of
        # these points
        points = np.random.rand(10, 3)
        rotation = rowan.random.rand(1)
        translation = np.random.rand(1, 3)
        transformed_points = rowan.rotate(rotation, points) + translation

        # Recover the rotation and check
        q, t = rowan.mapping.horn(points, transformed_points)

        assert np.logical_or(
            np.allclose(rotation, q), np.allclose(rotation, -q))
        assert np.allclose(translation, t)
    """
    X = np.atleast_2d(X)
    Y = np.atleast_2d(Y)
    if X.shape != Y.shape:
        raise ValueError("Input arrays must be the same shape")
    elif len(X.shape) != 2 or X.shape[1] != 3:
        raise ValueError("Input arrays must be (Nx3) arrays")

    # The algorithm depends on removing the centroid of the points.
    centroid_X = np.mean(X, axis=0)
    centroid_Y = np.mean(Y, axis=0)
    X_c = X - centroid_X
    Y_c = Y - centroid_Y

    A = np.empty((X.shape[0], 4, 4))
    A[:, 0, 0] = 0
    A[:, 1, 1] = 0
    A[:, 2, 2] = 0
    A[:, 3, 3] = 0
    A[:, [0, 3], [1, 2]] = -X_c[:, [0]]
    A[:, [1, 2], [0, 3]] = X_c[:, [0]]
    A[:, [0, 1], [2, 3]] = -X_c[:, [1]]
    A[:, [2, 3], [0, 1]] = X_c[:, [1]]
    A[:, [0, 2], [3, 1]] = -X_c[:, [2]]
    A[:, [1, 3], [2, 0]] = X_c[:, [2]]

    B = np.empty((Y.shape[0], 4, 4))
    B[:, 0, 0] = 0
    B[:, 1, 1] = 0
    B[:, 2, 2] = 0
    B[:, 3, 3] = 0
    B[:, [0, 2], [1, 3]] = -Y_c[:, [0]]
    B[:, [1, 3], [0, 2]] = Y_c[:, [0]]
    B[:, [0, 3], [2, 1]] = -Y_c[:, [1]]
    B[:, [2, 1], [0, 3]] = Y_c[:, [1]]
    B[:, [0, 1], [3, 2]] = -Y_c[:, [2]]
    B[:, [2, 3], [1, 0]] = Y_c[:, [2]]

    prods = np.matmul(A.transpose(0, 2, 1), B)
    N = np.sum(prods, axis=0)

    # Note that Horn advocates solving the characteristic polynomial
    # explicitly to avoid computing an eigendecomposition; we do so
    # for simplicity
    w, v = np.linalg.eig(N)
    q = v[:, np.argmax(w)]

    t = -rotate(q, centroid_X) + centroid_Y

    return q, t


def davenport(X, Y):
    R"""Find the optimal rotation and translation to map between two sets of
    points.

    This function implements the `Davenport q-method
    <https://ntrs.nasa.gov/search.jsp?R=19670009376>`_, the most robust method
    and basis of most modern solvers. It involves the construction of a
    particular matrix, the Davenport K-matrix, which is then diagonalized to
    find the appropriate eigenvalues. More modern algorithms aim to solve the
    characteristic equation directly rather than diagonalizing, which can
    provide speed benefits at the potential cost of robustness. The
    implementation in ``rowan`` does not do this, instead simply computing the
    spectral decomposition.

    Args:
        X ((N, 3) np.array): First set of N points.
        Y ((N, 3) np.array): Second set of N points.

    Returns:
        A tuple (q, t) where q is the quaternion to rotate the points and t
        is the translation.

    Example::

        import numpy as np

        # Create some random points, then make a random transformation of
        # these points
        points = np.random.rand(10, 3)
        rotation = rowan.random.rand(1)
        translation = np.random.rand(1, 3)
        transformed_points = rowan.rotate(rotation, points) + translation

        # Recover the rotation and check
        q, t = rowan.mapping.davenport(points, transformed_points)

        assert np.logical_or(
            np.allclose(rotation, q), np.allclose(rotation, -q))
        assert np.allclose(translation, t)
    """
    X = np.atleast_2d(X)
    Y = np.atleast_2d(Y)
    if X.shape != Y.shape:
        raise ValueError("Input arrays must be the same shape")
    elif len(X.shape) != 2 or X.shape[1] != 3:
        raise ValueError("Input arrays must be (Nx3) arrays")

    # The algorithm depends on removing the centroid of the points.
    centroid_X = np.mean(X, axis=0)
    centroid_Y = np.mean(Y, axis=0)
    X_c = X - centroid_X
    Y_c = Y - centroid_Y

    B = X_c.T.dot(Y_c)
    tr_B = np.trace(B)
    z = np.array([B[1, 2] - B[2, 1], B[2, 0] - B[0, 2], B[0, 1] - B[1, 0]]).T

    # Note that the original Davenport q-matrix puts the solution to the vector
    # part of the quaternion in the upper right block; this results in a
    # quaternion with scalar part in the 4th entry. The q-matrix here is
    # slightly modified to avoid this problem
    K = np.empty((4, 4))
    K[1:, 1:] = B + B.T - np.eye(3)*tr_B
    K[0, 1:] = z.T
    K[1:, 0] = z
    K[0, 0] = tr_B

    w, v = np.linalg.eig(K)
    q = v[:, np.argmax(w)]

    t = -rotate(q, centroid_X) + centroid_Y

    return q, t


def procrustes(X, Y, method='best', equivalent_quaternions=None):
    R"""Solve the orthogonal Procrustes problem with algorithmic options.

    Args:
        X ((N, m) np.array): First set of N points.
        Y ((N, m) np.array): Second set of N points.
        method (str): A method to use. Options are 'kabsch', 'davenport'
            and 'horn'. The default is to select the best option ('best').
        equivalent_quaternions (array-like): If the precise correspondence is
            not known, but the points are known to be part of a body with
            specific symmetries, the set of quaternions generating
            symmetry-equivalent configurations can be provided. These
            quaternions will be tested exhaustively to find the smallest
            symmetry-equivalent rotation.

    Returns:
        A tuple (q, t) where q is the quaternion to rotate the points and t
        is the translation.

    Example::

        import numpy as np

        # Create some random points, then make a random transformation of
        # these points
        points = np.random.rand(10, 3)
        rotation = rowan.random.rand(1)
        translation = np.random.rand(1, 3)
        transformed_points = rowan.rotate(rotation, points) + translation

        # Recover the rotation and check
        q, t = rowan.mapping.procrustes(
            points, transformed_points, method='horn')

        assert np.logical_or(
            np.allclose(rotation, q), np.allclose(rotation, -q))
        assert np.allclose(translation, t)
    """
    import sys
    thismodule = sys.modules[__name__]

    if method != 'best':
        try:
            method = getattr(thismodule, method)
        except KeyError:
            raise ValueError("The input method is not known")
    else:
        X = np.atleast_2d(X)
        Y = np.atleast_2d(Y)
        if X.shape != Y.shape:
            raise ValueError("Input arrays must be the same shape")
        elif len(X.shape) != 2:
            raise ValueError("Input arrays must be 2D arrays")
        if X.shape[1] != 3:
            method = getattr(thismodule, 'kabsch')
        else:
            method = getattr(thismodule, 'davenport')
    if equivalent_quaternions is not None:
        qs = []
        ts = []
        for eq in equivalent_quaternions:
            q, t = method(rotate(eq, X), Y)
            if method.__name__ == 'kabsch':
                qs.append(from_matrix(q))
            else:
                qs.append(q)
            ts.append(t)
        index = np.argmin([angle(q) for q in qs])
        return qs[index], ts[index]
    else:
        q, t = method(X, Y)
        if method == 'kabsch':
            return from_matrix(q), t
        else:
            return q, t


def icp(X, Y, method='best', unique_match=True, max_iterations=20,
        tolerance=0.001):
    R"""Find best mapping using the Iterative Closest Point algorithm.

    Args:
        X ((N, m) np.array): First set of N points.
        Y ((N, m) np.array): Second set of N points.
        method (str): A method to use for each alignment. Options are 'kabsch',
            'davenport' and 'horn'. The default is to select the best option
            ('best').
        unique_match (bool): Whether to require nearest neighbors to be unique.
        max_iterations (int): Number of iterations to attempt.
        tolerance (float): Indicates convergence.

    Returns:
        A tuple (R, t) where R is the matrix to rotate the points and t
        is the translation.

    Example::

        import numpy as np

        # Create some random points, then make a random transformation of
        # these points
        points = np.random.rand(10, 3)

        # Only works for small rotations
        rotation = rowan.from_axis_angle((1, 0, 0), 0.01)
        translation = np.random.rand(1, 3)
        transformed_points = rowan.rotate(rotation, points) + translation

        # Recover the rotation and check
        R, t = rowan.mapping.icp(points, transformed_points)
        q = rowan.from_matrix(R)

        assert np.logical_or(
            np.allclose(rotation, q), np.allclose(rotation, -q))
        assert np.allclose(translation, t)
    """

    import sys
    thismodule = sys.modules[__name__]

    if method != 'best':
        try:
            method = getattr(thismodule, method)
        except KeyError:
            raise ValueError("The input method is not known")
    else:
        X = np.atleast_2d(X)
        Y = np.atleast_2d(Y)
        if X.shape != Y.shape:
            raise ValueError("Input arrays must be the same shape")
        elif len(X.shape) != 2:
            raise ValueError("Input arrays must be (Nx3) arrays")
        if X.shape[1] != 3:
            method = getattr(thismodule, 'kabsch')
        else:
            method = getattr(thismodule, 'davenport')

    if unique_match:
        try:
            from scipy import spatial, optimize
        except ImportError:
            raise ImportError("Running with unique_match requires scipy. "
                              "Please install scipy and try again.")
    else:
        try:
            from sklearn import neighbors
            nn = neighbors.NearestNeighbors(n_neighbors=1)
            nn.fit(Y)
        except ImportError:
            raise ImportError("Running without unique_match requires "
                              "scikit-learn. Please install sklearn and try "
                              "again.")

    # Copy points so we have originals available.
    cur_points = np.copy(X)
    err_old = 0
    for i in range(max_iterations):
        # Rather than a coarse nearest neighbors, we apply the Hungarian
        # algorithm to ensure that we do not have duplicates. Unfortunately,
        # this precludes acceleration of the spatial search but is worthwhile
        # for the improved accuracy
        if unique_match:
            pair_distances = spatial.distance.cdist(cur_points, Y)
            row_ind, indices = optimize.linear_sum_assignment(pair_distances)
            distances = pair_distances[row_ind, indices]
        else:
            distances, indices = nn.kneighbors(cur_points, return_distance=True)
            distances = distances.ravel()
            indices = indices.ravel()

        # Compute current best transformation
        q, t = method(cur_points, Y[indices, :])

        # Update the current source
        if q.shape[-1] != 4:
            # Returned a matrix instead of a quaternion
            cur_points = np.dot(cur_points, q.T) + t
        else:
            cur_points = rotate(q, cur_points) + t

        # Tolerance check
        err = np.mean(distances)
        if np.abs(err_old - err) < tolerance:
            break
        err_old = err

    # calculate final transformation
    q, t = method(X, cur_points)

    if q.shape[-1] == 4:
        R = to_matrix(q)

    return R, t



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
R"""Submodule containing all standard functions"""
from __future__ import division, print_function, absolute_import

import numpy as np


def exp(q):
    R"""Computes the natural exponential function :math:`e^q`.

    The exponential of a quaternion in terms of its scalar and vector parts
    :math:`q = a + \boldsymbol{v}` is defined by exponential power series:
    formula :math:`e^x = \sum_{k=0}^{\infty} \frac{x^k}{k!}` as follows:

    .. math::
        \begin{align}
            e^q &= e^{a+v} \\
                &= e^a \left(\sum_{k=0}^{\infty} \frac{v^k}{k!} \right) \\
                &= e^a \left(\cos \lvert \lvert \boldsymbol{v} \rvert \rvert +
                    \frac{\boldsymbol{v}}{\lvert \lvert \boldsymbol{v} \rvert
                    \rvert} \sin \lvert \lvert \boldsymbol{v} \rvert \rvert
                    \right)
        \end{align}

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing exponentials of q.

    Example::

        q_exp = rowan.exp([1, 0, 0, 0])
    """
    # Ensure compatibility for numpy < 1.13; older numpy fail with
    # the fancy indexing used below when array is 1d
    q = np.asarray(q)
    if len(q.shape) == 1:
        flat = True
        q = np.atleast_2d(q)
    else:
        flat = False

    expo = np.empty(q.shape)
    norms = np.linalg.norm(q[..., 1:], axis=-1)
    e = np.exp(q[..., 0])
    expo[..., 0] = e * np.cos(norms)
    norm_zero = np.isclose(norms, 0)
    not_zero = np.logical_not(norm_zero)
    if np.any(not_zero):
        expo[not_zero, 1:] = e[not_zero, np.newaxis] * (
                q[not_zero, 1:]/norms[not_zero, np.newaxis]
                ) * np.sin(norms)[not_zero, np.newaxis]
        if np.any(norm_zero):
            expo[norm_zero, 1:] = 0
    else:
        expo[..., 1:] = 0

    if flat:
        return expo.squeeze()
    else:
        return expo


def expb(q, b):
    R"""Computes the exponential function :math:`b^q`.

    We define the exponential of a quaternion to an arbitrary base relative
    to the exponential function :math:`e^q` using the change of base
    formula as follows:

    .. math::
        \begin{align}
            b^q &= y \\
            q &= \log_b y  = \frac{\ln y}{\ln b}\\
            y &= e^{q\ln b}
        \end{align}

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing exponentials of q.

    Example::

        q_exp = rowan.expb([1, 0, 0, 0], 2)
    """
    q = np.asarray(q)
    return exp(q*np.log(b))


def exp10(q):
    R"""Computes the exponential function :math:`10^q`.

    Wrapper around :py:func:`expb`.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing exponentials of q.

    Example::

        q_exp = rowan.exp10([1, 0, 0, 0])
    """
    return expb(q, 10)


def log(q):
    R"""Computes the quaternion natural logarithm.

    The natural of a quaternion in terms of its scalar and vector parts
    :math:`q = a + \boldsymbol{v}` is defined by inverting the exponential
    formula (see :py:func:`exp`), and is defined by the formula
    :math:`\frac{x^k}{k!}` as follows:

    .. math::
        \begin{equation}
            \ln(q) = \ln\lvert\lvert q \rvert\rvert +
                    \frac{\boldsymbol{v}}{\lvert\lvert \boldsymbol{v}
                    \rvert\rvert} \arccos\left(\frac{a}{q}\right)
        \end{equation}

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing logarithms of q.

    Example::

        ln_q  = rowan.log([1, 0, 0, 0])
    """
    # Ensure compatibility for numpy < 1.13; older numpy fail with
    # the fancy indexing used below when array is 1d
    q = np.asarray(q)
    if len(q.shape) == 1:
        flat = True
        q = np.atleast_2d(q)
    else:
        flat = False

    log = np.empty(q.shape)

    # We need all the norms to avoid divide by zeros later.
    # Can also use these to minimize the amount of work done.
    q_norms = norm(q)
    q_norm_zero = np.isclose(q_norms, 0)
    q_not_zero = np.logical_not(q_norm_zero)
    v_norms = np.linalg.norm(q[..., 1:], axis=-1)
    v_norm_zero = np.isclose(v_norms, 0)
    v_not_zero = np.logical_not(v_norm_zero)

    if np.any(q_not_zero):
        if np.any(q_norm_zero):
            log[q_norm_zero, 0] = -np.inf
        log[q_not_zero, 0] = np.log(q_norms[q_not_zero])
    else:
        log[..., 0] = -np.inf

    if np.any(v_not_zero):
        prefactor = np.empty(q[v_not_zero, 1:].shape)
        prefactor = q[v_not_zero, 1:]/v_norms[
                v_not_zero, np.newaxis]

        inv_cos = np.empty(v_norms[v_not_zero].shape)
        inv_cos = np.arccos(q[v_not_zero, 0]/q_norms[v_not_zero])

        if np.any(v_norm_zero):
            log[v_norm_zero, 1:] = 0
        log[v_not_zero, 1:] = prefactor * inv_cos[..., np.newaxis]
    else:
        log[..., 1:] = 0

    if flat:
        return log.squeeze()
    else:
        return log


def logb(q, b):
    R"""Computes the quaternion logarithm to some base b.

    The quaternion logarithm for arbitrary bases is defined using the
    standard change of basis formula relative to the natural logarithm.

    .. math::
        \begin{align}
            \log_b q &= y \\
            q &= b^y \\
            \ln q &= y \ln b \\
            y &= \log_b q = \frac{\ln q}{\ln b}
        \end{align}

    Args:
        q ((...,4) np.array): Array of quaternions.
        n ((...) np.array): Scalars to use as log bases.

    Returns:
        Array of shape (...) containing logarithms of q.

    Example::

        log2_q = rowan.logb([1, 0, 0, 0], 2)
    """
    q = np.asarray(q)
    return log(q)/np.log(b)


def log10(q):
    R"""Computes the quaternion logarithm base 10.

    Wrapper around :py:func:`logb`.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing logarithms of q.

    Example::

        log10_q = rowan.log10([1, 0, 0, 0])
    """
    q = np.asarray(q)
    return logb(q, 10)


def power(q, n):
    R"""Computes the power of a quaternion :math:`q^n`.

    Quaternions raised to a scalar power are defined according to the polar
    decomposition angle :math:`\theta` and vector :math:`\hat{u}`:
    :math:`q^n = \lvert\lvert q \rvert\rvert^n \left( \cos(n\theta) + \hat{u}
    \sin(n\theta)\right)`. However, this can be computed
    more efficiently by noting that :math:`q^n = \exp(n \ln(q))`.

    Args:
        q ((...,4) np.array): Array of quaternions.
        n ((...) np.arrray): Scalars to exponentiate quaternions with.

    Returns:
        Array of shape (...) containing powers of q.

    Example::

        q_5 = rowan.power([1, 0, 0, 0], 5)
    """
    # TODO: Write polar decomposition function #noqa
    q = np.asarray(q)
    # Need matching shapes
    if len(q.shape) == 1:
        flat = True
        q = np.atleast_2d(q)
    else:
        flat = False

    newshape = np.broadcast(q[..., 0], n).shape
    q = np.broadcast_to(q, newshape + (4,))
    n = np.broadcast_to(n, newshape)

    # Note that we follow the convention that 0^0 = 1
    check = (n == 0)
    if np.any(check):
        powers = np.empty(newshape + (4,))
        powers[check] = np.array([1, 0, 0, 0])
        not_check = np.logical_not(check)
        if np.any(not_check):
            powers[not_check] = exp(
                    n[not_check, np.newaxis] * log(q[not_check, :]))
    else:
        powers = exp(n[..., np.newaxis]*log(q))

    if flat:
        return powers.squeeze()
    else:
        return powers


def conjugate(q):
    R"""Conjugates an array of quaternions.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing conjugates of q.

    Example::

        q_star = rowan.conjugate([1, 0, 0, 0])
    """
    # Don't use asarray to avoid modifying in place
    conjugate = np.array(q)
    conjugate[..., 1:] *= -1
    return conjugate


def inverse(q):
    R"""Computes the inverse of an array of quaternions.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing inverses of q.

    Example::

        q_inv = rowan.inverse([1, 0, 0, 0])
    """
    q = np.asarray(q)

    if len(q.shape) == 1:
        flat = True
        inverses = np.array(np.atleast_2d(q))
    else:
        flat = False
        inverses = np.array(q)

    normsq = norm(inverses)**2
    if np.any(normsq):
        inverses[..., 1:] *= -1
        # Would like to do this in place, but can't guarantee type safety
        inverses[normsq > 0] = inverses[normsq > 0]/normsq[
                normsq > 0, np.newaxis]

    if flat:
        return inverses.squeeze()
    else:
        return inverses


def multiply(qi, qj):
    R"""Multiplies two arrays of quaternions.

    Note that quaternion multiplication is generally non-commutative, so the
    first and second set of quaternions must be passed in the correct order.

    Args:
        qi ((...,4) np.array): Array of left quaternions.
        qj ((...,4) np.array): Array of right quaternions.

    Returns:
        Array of shape (...) containing element-wise products of q.

    Example::

        prod = rowan.multiply([1, 0, 0, 0], [2, 0, 0, 0])
    """
    qi = np.asarray(qi)
    qj = np.asarray(qj)

    output = np.empty(np.broadcast(qi, qj).shape)

    output[..., 0] = qi[..., 0] * qj[..., 0] - \
        np.sum(qi[..., 1:] * qj[..., 1:], axis=-1)
    output[..., 1:] = (qi[..., 0, np.newaxis] * qj[..., 1:] +
                       qj[..., 0, np.newaxis] * qi[..., 1:] +
                       np.cross(qi[..., 1:], qj[..., 1:]))
    return output


def divide(qi, qj):
    R"""Divides two arrays of quaternions.

    Division is non-commutative; this function returns
    :math:`q_i q_j^{-1}`.

    Args:
        qi ((...,4) np.array): Dividend quaternions.
        qj ((...,4) np.array): Divisor quaternions.

    Returns:
        Array of shape (...) containing element-wise quotients of qi and qj.

    Example::

        quot = rowan.divide([1, 0, 0, 0], [2, 0, 0, 0])
    """
    return multiply(qi, inverse(qj))


def norm(q):
    R"""Compute the quaternion norm.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) containing norms of q.

    Example::

        norms = rowan.norm([10, 0, 0, 0])
    """
    q = np.asarray(q)
    return np.linalg.norm(q, axis=-1)


def normalize(q):
    R"""Normalize quaternions.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        Array of shape (...) of normalized quaternions.

    Example::

        u = rowan.normalize([10, 0, 0, 0])
    """
    q = np.asarray(q)
    norms = norm(q)
    return q / norms[..., np.newaxis]


def is_unit(q):
    """Check if all input quaternions have unit norm.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        bool: Whether or not all inputs are unit quaternions

    Example::

        rowan.is_unit([10, 0, 0, 0])
   """
    return np.allclose(norm(q), 1)


def _validate_unit(q, msg="Arguments must be unit quaternions"):
    """Simple helper function to ensure that all quaternions in q are unit."""
    if not is_unit(q):
        raise ValueError(msg)


def from_mirror_plane(x, y, z):
    R"""Generate quaternions from mirror plane equations.

    Reflection quaternions can be constructed from the form
    :math:`(0, x, y, z)`, *i.e.* with zero real component. The vector
    :math:`(x, y, z)` is the normal to the mirror plane.

    Args:
        x ((...) np.array): First planar component.
        y ((...) np.array): Second planar component.
        z ((...) np.array): Third planar component.

    Returns:
        Array of shape (...) containing quaternions reflecting about the input
        plane :math:`(x, y, z)`.

    Example::

        quat_ref = rowan.from_mirror_plane(*(1, 2, 3))
    """
    x, y, z = np.broadcast_arrays(x, y, z)
    q = np.empty(x.shape + (4,))
    q[..., 0] = 0
    q[..., 1] = x
    q[..., 2] = y
    q[..., 3] = z

    return q


def _promote_vec(v):
    R"""Helper function to promote vectors to their quaternion representation.
    """
    return np.concatenate((np.zeros(v.shape[:-1] + (1,)), v), axis=-1)


def reflect(q, v):
    R"""Reflect a list of vectors by a corresponding set of quaternions.

    For help constructing a mirror plane, see :py:func:`from_mirror_plane`.

    Args:
        q ((...,4) np.array): Array of quaternions.
        v ((...,3) np.array): Array of vectors.

    Returns:
        Array of shape (..., 3) containing reflections of v.

    Example::

        v_reflected = rowan.reflect([1, 0, 0, 0], [1, 1, 1])
    """
    q = np.asarray(q)
    v = np.asarray(v)

    # Convert vector to quaternion representation
    quat_v = _promote_vec(v)
    return multiply(q, multiply(quat_v, q))[..., 1:]


def rotate(q, v):
    R"""Rotate a list of vectors by a corresponding set of quaternions.

    Args:
        q ((...,4) np.array): Array of quaternions.
        v ((...,3) np.array): Array of vectors.

    Returns:
        Array of shape (..., 3) containing rotations of v.

    Example::

        v_rot = rowan.reflect([1, 0, 0, 0], [1, 1, 1])
    """
    q = np.asarray(q)
    v = np.asarray(v)

    # Convert vector to quaternion representation
    quat_v = _promote_vec(v)
    return multiply(q, multiply(quat_v, conjugate(q)))[..., 1:]


def _normalize_vec(v):
    R"""Helper function to normalize vectors."""
    v = np.asarray(v)
    norms = np.linalg.norm(v, axis=-1)
    return v / norms[..., np.newaxis]


def _vector_bisector(v1, v2):
    R"""Find the vector bisecting two vectors.

    Args:
        v1 ((...,3) np.array): First array of vectors.
        v2 ((...,3) np.array): Second array of vectors.

    Returns:
        Array of shape (..., 3) containing vector bisectors.
    """
    # Since np.inner and np.dot require manipulating the shapes in ways that
    # might be expensive and may not play nicely with broadcasting, we perform
    # the dot product manually on the broadcasted arrays
    v1_norm, v2_norm = np.broadcast_arrays(_normalize_vec(v1),
                                           _normalize_vec(v2))
    ap = np.isclose(np.sum(v1_norm*v2_norm, axis=-1), -1)

    if np.any(ap):
        result = np.empty(v1_norm.shape)

        # Parallel vectors are fine, only antiparallel vectors cause problems
        not_ap = np.logical_not(ap)
        result[not_ap] = _normalize_vec(v1_norm[not_ap] + v2_norm[not_ap])

        # To use cross products to find the normal, we need to choose a unit
        # vector that is also not (anti)parallel to the original. Keep two
        # options available to avoid this case.
        one_vec = np.array([[1, 0, 0]])
        other_one_vec = np.array([[0, 1, 0]])
        cross_element = np.where(
                            np.isclose(
                                np.abs(np.dot(v1_norm[ap], one_vec.T)),
                                1),
                            other_one_vec,
                            one_vec)
        result[ap] = np.cross(v1_norm[ap], cross_element)

        return result
    else:
        return _normalize_vec(v1_norm + v2_norm)


def vector_vector_rotation(v1, v2):
    R"""Find the quaternion to rotate one vector onto another.

    Args:
        v1 ((...,3) np.array): Array of vectors to rotate.
        v2 ((...,3) np.array): Array of vector to rotate onto.

    Returns:
        Array of shape (..., 4) containing  quaternions that rotate v1 onto v2.

    Example::

        q_rot = rowan.vector_vector_rotation([1, 0, 0], [0, 1, 0])
    """
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    return from_axis_angle(_vector_bisector(v1, v2), np.pi)


def from_euler(alpha, beta, gamma, convention='zyx',
               axis_type='intrinsic'):
    R"""Convert Euler angles to quaternions.

    For generality, the rotations are computed by composing a sequence of
    quaternions corresponding to axis-angle rotations. While more efficient
    implementations are possible, this method was chosen to prioritize
    flexibility since it works for essentially arbitrary Euler angles as
    long as intrinsic and extrinsic rotations are not intermixed.

    Args:
        alpha ((...) np.array): Array of :math:`\alpha` values in radians.
        beta ((...) np.array): Array of :math:`\beta` values in radians.
        gamma ((...) np.array): Array of :math:`\gamma` values in radians.
        convention (str): One of the 12 valid conventions xzx, xyx,
            yxy, yzy, zyz, zxz, xzy, xyz, yxz, yzx, zyx, zxy.
        axes (str): Whether to use extrinsic or intrinsic rotations.

    Returns:
        Array of shape (..., 4) containing quaternions corresponding to the
        input angles.

    Example::

        ql = rowan.from_euler(0.3, 0.5, 0.7)
    """
    angles = np.broadcast_arrays(alpha, beta, gamma)

    convention = convention.lower()

    if len(convention) > 3 or (set(convention) - set('xyz')):
        raise ValueError("All acceptable conventions must be 3 \
character strings composed only of x, y, and z")

    basis_axes = {
        'x': np.array([1, 0, 0]),
        'y': np.array([0, 1, 0]),
        'z': np.array([0, 0, 1]),
    }
    # Temporary method to ensure shapes conform
    for ax, vec in basis_axes.items():
        basis_axes[ax] = np.broadcast_to(
            vec,
            angles[0].shape + (vec.shape[-1],)
        )

    # Split by convention, the easiest
    rotations = []
    if axis_type == 'extrinsic':
        # Loop over the axes and add each rotation
        for i, char in enumerate(convention):
            ax = basis_axes[char]
            rotations.append(from_axis_angle(ax, angles[i]))
    elif axis_type == 'intrinsic':
        for i, char in enumerate(convention):
            ax = basis_axes[char]
            rotations.append(from_axis_angle(ax, angles[i]))
            # Rotate the bases as well
            for key, value in basis_axes.items():
                basis_axes[key] = rotate(
                    rotations[-1],
                    value
                )
    else:
        raise ValueError("Only valid axis_types are intrinsic and extrinsic")

    # Compose the total rotation
    final_rotation = np.broadcast_to(
        np.array([1, 0, 0, 0]),
        rotations[0].shape
    )
    for q in rotations:
        final_rotation = multiply(q, final_rotation)

    return final_rotation


def to_euler(q, convention='zyx', axis_type='intrinsic'):
    R"""Convert quaternions to Euler angles.

    Euler angles are returned in the sequence provided, so in, *e.g.*,
    the default case ('zyx'), the angles returned are for a rotation
    :math:`Z(\alpha) Y(\beta) X(\gamma)`.

    .. note::

        In all cases, the :math:`\alpha` and :math:`\gamma` angles are
        between :math:`\pm \pi`. For proper Euler angles, :math:`\beta`
        is between :math:`0` and :math:`pi` degrees. For Tait-Bryan
        angles, :math:`\beta` lies between :math:`\pm\pi/2`.

    For simplicity, quaternions are converted to matrices, which are
    then converted to their Euler angle representations. All equations
    for rotations are derived by considering compositions of the three
    elemental rotations about the three Cartesian axes:

    .. math::
        :nowrap:

        \begin{eqnarray*}
        R_x(\theta)  =& \left(\begin{array}{ccc}
                            1 & 0           & 0            \\
                            0 & \cos \theta & -\sin \theta \\
                            0 & \sin \theta & \cos \theta  \\
                         \end{array}\right)\\
        R_y(\theta)  =& \left(\begin{array}{ccc}
                            \cos \theta  & 0 & \sin \theta \\
                            0            & 1 &  0          \\
                            -\sin \theta & 1 & \cos \theta \\
                         \end{array}\right)\\
        R_z(\theta)  =& \left(\begin{array}{ccc}
                            \cos \theta & -\sin \theta & 0 \\
                            \sin \theta & \cos \theta  & 0 \\
                            0           & 0            & 1 \\
                         \end{array}\right)\\
        \end{eqnarray*}

    Extrinsic rotations are represented by matrix multiplications in
    the proper order, so :math:`z-y-x` is represented by the
    multiplication :math:`XYZ` so that the system is rotated first
    about :math:`Z`, then about :math:`Y`, then finally :math:`X`.
    For intrinsic rotations, the order of rotations is reversed,
    meaning that it matches the order in which the matrices actually
    appear *i.e.* the :math:`z-y'-x''` convention (yaw, pitch, roll)
    corresponds to the multiplication of matrices :math:`ZYX`.
    For proof of the relationship between intrinsic and extrinsic
    rotations, see the `Wikipedia page on Davenport chained rotations
    <https://en.wikipedia.org/wiki/Davenport_chained_rotations>`_.

    For more information, see the Wikipedia page for
    `Euler angles <https://en.wikipedia.org/wiki/Euler_angles>`_
    (specifically the section on converting between representations).

    Args:
        q ((...,4) np.array): Quaternions to transform.
        convention (str): One of the 6 valid conventions zxz,
            xyx, yzy, zyz, xzx, yxy.
        axes (str): Whether to use extrinsic or intrinsic.

    Returns:
        Array of shape (..., 3) containing Euler angles :math:`(\alpha, \beta,
        \gamma)` as the last dimension (in radians).

    Example::

        import numpy as np
        rands = np.random.rand(100, 3)
        alpha, beta, gamma = rands.T
        ql = rowan.from_euler(alpha, beta, gamma)
        alpha_return, beta_return, gamma_return = np.split(
            rowan.to_euler(ql), 3, axis = 1)
        assert(np.allclose(alpha_return.flatten(), alpha))
        assert(np.allclose(beta_return.flatten(), beta))
        assert(np.allclose(gamma_return.flatten(), gamma))
    """
    q = np.asarray(q)
    _validate_unit(q)

    try:
        mats = to_matrix(q)
    except ValueError:
        raise ValueError(
            "Not all quaternions in q are unit quaternions.")

    if axis_type == 'intrinsic':
        # Have to hardcode the different possibilities.
        # Classical Euler angles
        if convention == 'xzx':
            alpha = np.arctan2(mats[..., 2, 0], mats[..., 1, 0])
            beta = np.arccos(mats[..., 0, 0])
            gamma = np.arctan2(mats[..., 0, 2], -mats[..., 0, 1])
        elif convention == 'xyx':
            alpha = np.arctan2(mats[..., 1, 0], -mats[..., 2, 0])
            beta = np.arccos(mats[..., 0, 0])
            gamma = np.arctan2(mats[..., 0, 1], mats[..., 0, 2])
        elif convention == 'yxy':
            alpha = np.arctan2(mats[..., 0, 1], mats[..., 2, 1])
            beta = np.arccos(mats[..., 1, 1])
            gamma = np.arctan2(mats[..., 1, 0], -mats[..., 1, 2])
        elif convention == 'yzy':
            alpha = np.arctan2(mats[..., 2, 1], -mats[..., 0, 1])
            beta = np.arccos(mats[..., 1, 1])
            gamma = np.arctan2(mats[..., 1, 2], mats[..., 1, 0])
        elif convention == 'zyz':
            alpha = np.arctan2(mats[..., 1, 2], mats[..., 0, 2])
            beta = np.arccos(mats[..., 2, 2])
            gamma = np.arctan2(mats[..., 2, 1], -mats[..., 2, 0])
        elif convention == 'zxz':
            alpha = np.arctan2(mats[..., 0, 2], -mats[..., 1, 2])
            beta = np.arccos(mats[..., 2, 2])
            gamma = np.arctan2(mats[..., 2, 0], mats[..., 2, 1])
        # Tait-Bryan angles
        elif convention == 'xzy':
            alpha = np.arctan2(mats[..., 2, 1], mats[..., 1, 1])
            beta = np.arcsin(-mats[..., 0, 1])
            gamma = np.arctan2(mats[..., 0, 2], mats[..., 0, 0])
        elif convention == 'xyz':
            alpha = np.arctan2(-mats[..., 1, 2], mats[..., 2, 2])
            beta = np.arcsin(mats[..., 0, 2])
            gamma = np.arctan2(-mats[..., 0, 1], mats[..., 0, 0])
        elif convention == 'yxz':
            alpha = np.arctan2(mats[..., 0, 2], mats[..., 2, 2])
            beta = np.arcsin(-mats[..., 1, 2])
            gamma = np.arctan2(mats[..., 1, 0], mats[..., 1, 1])
        elif convention == 'yzx':
            alpha = np.arctan2(-mats[..., 2, 0], mats[..., 0, 0])
            beta = np.arcsin(mats[..., 1, 0])
            gamma = np.arctan2(-mats[..., 1, 2], mats[..., 1, 1])
        elif convention == 'zyx':
            alpha = np.arctan2(mats[..., 1, 0], mats[..., 0, 0])
            beta = np.arcsin(-mats[..., 2, 0])
            gamma = np.arctan2(mats[..., 2, 1], mats[..., 2, 2])
        elif convention == 'zxy':
            alpha = np.arctan2(-mats[..., 0, 1], mats[..., 1, 1])
            beta = np.arcsin(mats[..., 2, 1])
            gamma = np.arctan2(-mats[..., 2, 0], mats[..., 2, 2])
        else:
            raise ValueError("Unknown convention selected!")
    elif axis_type == 'extrinsic':
        # For these, the matrix must be constructed in reverse order
        # e.g. Z(\alpha)Y'(\beta)Z''(\gamma) (where primes denote the
        # rotated frames) becomes the extrinsic rotation
        # Z(\gamma)Y(\beta)Z(\alpha).

        # Classical Euler angles
        if convention == 'xzx':
            alpha = np.arctan2(mats[..., 0, 2], -mats[..., 0, 1])
            beta = np.arccos(mats[..., 0, 0])
            gamma = np.arctan2(mats[..., 2, 0], mats[..., 1, 0])
        elif convention == 'xyx':
            alpha = np.arctan2(mats[..., 0, 1], mats[..., 0, 2])
            beta = np.arccos(mats[..., 0, 0])
            gamma = np.arctan2(mats[..., 1, 0], -mats[..., 2, 0])
        elif convention == 'yxy':
            alpha = np.arctan2(mats[..., 1, 0], -mats[..., 1, 2])
            beta = np.arccos(mats[..., 1, 1])
            gamma = np.arctan2(mats[..., 0, 1], mats[..., 2, 1])
        elif convention == 'yzy':
            alpha = np.arctan2(mats[..., 1, 2], mats[..., 1, 0])
            beta = np.arccos(mats[..., 1, 1])
            gamma = np.arctan2(mats[..., 2, 1], -mats[..., 0, 1])
        elif convention == 'zyz':
            alpha = np.arctan2(mats[..., 2, 1], -mats[..., 2, 0])
            beta = np.arccos(mats[..., 2, 2])
            gamma = np.arctan2(mats[..., 1, 2], mats[..., 0, 2])
        elif convention == 'zxz':
            alpha = np.arctan2(mats[..., 2, 0], mats[..., 2, 1])
            beta = np.arccos(mats[..., 2, 2])
            gamma = np.arctan2(mats[..., 0, 2], -mats[..., 1, 2])
        # Tait-Bryan angles
        elif convention == 'xzy':
            alpha = np.arctan2(-mats[..., 1, 2], mats[..., 1, 1])
            beta = np.arcsin(mats[..., 1, 0])
            gamma = np.arctan2(-mats[..., 2, 0], mats[..., 0, 0])
        elif convention == 'xyz':
            alpha = np.arctan2(mats[..., 2, 1], mats[..., 2, 2])
            beta = np.arcsin(-mats[..., 2, 0])
            gamma = np.arctan2(mats[..., 1, 0], mats[..., 0, 0])
        elif convention == 'yxz':
            alpha = np.arctan2(-mats[..., 2, 0], mats[..., 2, 2])
            beta = np.arcsin(mats[..., 2, 1])
            gamma = np.arctan2(-mats[..., 0, 1], mats[..., 1, 1])
        elif convention == 'yzx':
            alpha = np.arctan2(mats[..., 0, 2], mats[..., 0, 0])
            beta = np.arcsin(-mats[..., 0, 1])
            gamma = np.arctan2(mats[..., 2, 1], mats[..., 1, 1])
        elif convention == 'zyx':
            alpha = np.arctan2(-mats[..., 0, 1], mats[..., 0, 0])
            beta = np.arcsin(mats[..., 0, 2])
            gamma = np.arctan2(-mats[..., 1, 2], mats[..., 2, 2])
        elif convention == 'zxy':
            alpha = np.arctan2(mats[..., 1, 0], mats[..., 1, 1])
            beta = np.arcsin(-mats[..., 1, 2])
            gamma = np.arctan2(mats[..., 0, 2], mats[..., 2, 2])
        else:
            raise ValueError("Unknown convention selected!")
    else:
        raise ValueError("The axis type must be either extrinsic or intrinsic")

    return np.stack((alpha, beta, gamma), axis=-1)


def from_matrix(mat, require_orthogonal=True):
    R"""Convert the rotation matrices mat to quaternions.

    This method uses the algorithm described by Bar-Itzhack in [Itzhack00]_.
    The idea is to construct a matrix K whose largest eigenvalue corresponds
    to the desired quaternion. One of the strengths of the algorithm is that
    for nonorthogonal matrices it gives the closest quaternion representation
    rather than failing outright.

    .. [Itzhack00] Itzhack Y. Bar-Itzhack.  "New Method for Extracting the
        Quaternion from a Rotation Matrix", Journal of Guidance, Control, and
        Dynamics, Vol. 23, No. 6 (2000), pp. 1085-1087
        https://doi.org/10.2514/2.4654

    Args:
        mat ((...,3,3) np.array): An array of rotation matrices.

    Returns:
        Array of shape (..., 4) containing the corresponding rotation
        quaternions.

    Example::

        ql = rowan.from_matrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    """
    mat = np.asarray(mat)
    if require_orthogonal and not np.allclose(np.linalg.det(mat), 1):
        raise ValueError(
            "Not all of your matrices are orthogonal. \
Please ensure that there are no improper rotations. \
If this was intentional, set require_orthogonal to \
False when calling this function.")

    K = np.zeros(mat.shape[:-2] + (4, 4))
    K[..., 0, 0] = mat[..., 0, 0] - mat[..., 1, 1] - mat[..., 2, 2]
    K[..., 0, 1] = mat[..., 1, 0] + mat[..., 0, 1]
    K[..., 0, 2] = mat[..., 2, 0] + mat[..., 0, 2]
    K[..., 0, 3] = mat[..., 1, 2] - mat[..., 2, 1]
    K[..., 1, 0] = mat[..., 1, 0] + mat[..., 0, 1]
    K[..., 1, 1] = mat[..., 1, 1] - mat[..., 0, 0] - mat[..., 2, 2]
    K[..., 1, 2] = mat[..., 2, 1] + mat[..., 1, 2]
    K[..., 1, 3] = mat[..., 2, 0] - mat[..., 0, 2]
    K[..., 2, 0] = mat[..., 2, 0] + mat[..., 0, 2]
    K[..., 2, 1] = mat[..., 2, 1] + mat[..., 1, 2]
    K[..., 2, 2] = mat[..., 2, 2] - mat[..., 0, 0] - mat[..., 1, 1]
    K[..., 2, 3] = mat[..., 0, 1] - mat[..., 1, 0]
    K[..., 3, 0] = mat[..., 1, 2] - mat[..., 2, 1]
    K[..., 3, 1] = mat[..., 2, 0] - mat[..., 0, 2]
    K[..., 3, 2] = mat[..., 0, 1] - mat[..., 1, 0]
    K[..., 3, 3] = mat[..., 0, 0] + mat[..., 1, 1] + mat[..., 2, 2]
    K = K / 3.0

    _, v = np.linalg.eigh(K)
    # The conventions in the paper are very confusing for quaternions in terms
    # of the order of the components
    return np.concatenate(
        (v[..., -1, -1, np.newaxis], -v[..., :-1, -1]), axis=-1)


def to_matrix(q, require_unit=True):
    R"""Convert quaternions into rotation matrices.

    Uses the conversion described on `Wikipedia
    <https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Quaternion-derived_rotation_matrix>`_.

    Args:
        q ((...,4) np.array): An array of quaternions.

    Returns:
        Array of shape (..., 3, 3) containing the corresponding rotation
        matrices.

    Example::

        ql = rowan.to_matrix([1, 0, 0, 0]])
    """
    q = np.asarray(q)

    s = norm(q)
    if np.any(s == 0.0):
        raise ZeroDivisionError(
            "At least one element of q has approximately zero norm")
    elif require_unit and not np.allclose(s, 1.0):
        raise ValueError(
            "Not all quaternions in q are unit quaternions. \
If this was intentional, please set require_unit to False when \
calling this function.")
    m = np.empty(q.shape[:-1] + (3, 3))
    s **= -1.0  # For consistency with Wikipedia notation
    m[..., 0, 0] = 1.0 - 2 * s * (q[..., 2]**2 + q[..., 3]**2)
    m[..., 0, 1] = 2 * (q[..., 1] * q[..., 2] - q[..., 3] * q[..., 0])
    m[..., 0, 2] = 2 * (q[..., 1] * q[..., 3] + q[..., 2] * q[..., 0])
    m[..., 1, 0] = 2 * (q[..., 1] * q[..., 2] + q[..., 3] * q[..., 0])
    m[..., 1, 1] = 1.0 - 2 * (q[..., 1]**2 + q[..., 3]**2)
    m[..., 1, 2] = 2 * (q[..., 2] * q[..., 3] - q[..., 1] * q[..., 0])
    m[..., 2, 0] = 2 * (q[..., 1] * q[..., 3] - q[..., 2] * q[..., 0])
    m[..., 2, 1] = 2 * (q[..., 2] * q[..., 3] + q[..., 1] * q[..., 0])
    m[..., 2, 2] = 1.0 - 2 * (q[..., 1]**2 + q[..., 2]**2)
    return m


def from_axis_angle(axes, angles):
    R"""Find quaternions to rotate a specified angle about a specified axis.

    Args:
        axes ((...,3) np.array): An array of vectors (the axes).
        angles (float or (...,1) np.array): An array of angles in radians.
            Will be broadcast to match shape of v as needed.

    Returns:
        Array of shape (..., 4) containing the corresponding rotation
        quaternions.

    Example::

        quat = rowan.from_axis_angle([[1, 0, 0]], np.pi/3)
    """
    axes = np.asarray(axes)

    # First reshape angles and compute the half angle
    bc = np.broadcast(angles, axes[..., 0])
    angles = np.broadcast_to(angles, bc.shape)[..., np.newaxis]
    axes = np.broadcast_to(axes, bc.shape + (3,))
    ha = angles / 2.0

    # Normalize the vector
    u = _normalize_vec(axes)

    # Compute the components of the quaternions
    scalar_comp = np.cos(ha)
    vec_comp = np.sin(ha) * u

    return np.concatenate((scalar_comp, vec_comp), axis=-1)


def to_axis_angle(q):
    R"""Convert the quaternions in q to axis angle representations.

    Args:
        q ((...,4) np.array): An array of quaternions.

    Returns:
        A tuple of np.arrays (axes, angles) where axes has
        shape (...,3) and angles has shape (...,1). The
        angles are in radians.

    Example::

        quat = rowan.to_axis_angle([[1, 0, 0, 0]])
    """
    q = np.asarray(q)
    _validate_unit(q)

    angles = 2*np.atleast_1d(np.arccos(q[..., 0]))
    sines = np.sin(angles/2)
    # Avoid divide by zero issues; these values will not be used
    sines[sines == 0] = 1
    axes = np.where(angles != 0,
                    q[..., 1:]/sines,
                    0)

    return axes, angles


def equal(p, q):
    R"""Check whether two sets of quaternions are equal.

    This function is a simple wrapper that checks array
    equality and then aggregates along the quaternion axis.

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.

    Returns:
        A boolean array of shape (...) indicating equality.

    Example::

        rowan.equal([1, 0, 0, 0], [1, 0, 0, 0])
    """
    return np.all(p == q, axis=-1)


def not_equal(p, q):
    R"""Check whether two sets of quaternions are not equal.

    This function is a simple wrapper that checks array
    equality and then aggregates along the quaternion axis.

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.

    Returns:
        A boolean array of shape (...) indicating inequality.

    Example::

        rowan.not_equal([-1, 0, 0, 0], [1, 0, 0, 0])
    """
    return np.any(p != q, axis=-1)


def isnan(q):
    R"""Test element-wise for NaN quaternions.

    A quaternion is defined as NaN if any elements are NaN.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        A boolean array of shape (...) indicating whether or not the input
        quaternions were NaN.

    Example::

        import numpy as np
        rowan.isnan([np.nan, 0, 0, 0])
    """
    return np.any(np.isnan(q), axis=-1)


def isinf(q):
    R"""Test element-wise for infinite quaternions.

    A quaternion is defined as infinite if any elements are infinite.

    Args:
        q ((...,4) np.array): Array of quaternions

    Returns:
        A boolean array of shape (...) indicating infinite quaternions.

    Example::

        import numpy as np
        rowan.isinf([np.nan, 0, 0, 0])
    """
    return np.any(np.isinf(q), axis=-1)


def isfinite(q):
    R"""Test element-wise for finite quaternions.

    A quaternion is defined as finite if all elements are finite.

    Args:
        q ((...,4) np.array): Array of quaternions.

    Returns:
        A boolean array of shape (...) indicating finite quaternions.

    Example::

        rowan.isfinite([1, 0, 0, 0])
    """
    return np.all(np.isfinite(q), axis=-1)


def allclose(p, q, **kwargs):
    R"""Check whether two sets of quaternions are all close.

    This is a direct wrapper of the corresponding NumPy function.

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.
        **kwargs: Keyword arguments to pass to np.allclose.

    Returns:
        Boolean indicating whether or not all quaternions are close.

    Example::

        rowan.allclose([1, 0, 0, 0], [1, 0, 0, 0])
    """
    return np.allclose(p, q, **kwargs)


def isclose(p, q, **kwargs):
    R"""Element-wise check of whether two sets of quaternions are close.

    This function is a simple wrapper that checks using the
    corresponding NumPy function and then aggregates along
    the quaternion axis.

    Args:
        p ((...,4) np.array): First array of quaternions.
        q ((...,4) np.array): Second array of quaternions.
        **kwargs: Keyword arguments to pass to np.isclose.

    Returns:
        A boolean array of shape (...) indicating which quaternions are close.

    Example::

        rowan.allclose([[1, 0, 0, 0]], [[1, 0, 0, 0]])
    """
    return np.all(np.isclose(p, q, **kwargs), axis=-1)



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
R"""This subpackage provides the ability to compute the derivative and
integral of a quaternion.
"""
from __future__ import division, print_function, absolute_import

import numpy as np

from ..functions import multiply, _promote_vec, _validate_unit, exp

__all__ = ['derivative',
           'integrate']


def derivative(q, v):
    R"""Compute the instantaneous derivative of unit quaternions, which is
    defined as

    .. math::
        \dot{q} = \frac{1}{2} \boldsymbol{v} q

    A derivation is provided `here`_.
    For a more thorough explanation, see `this page`_.

    .. _here: http://web.cs.iastate.edu/~cs577/handouts/quaternion.pdf
    .. _this page: https://fgiesen.wordpress.com/
                   2012/08/24/quaternion-differentiation/

    Args:
        q ((...,4) np.array): Array of quaternions.
        v ((...,3) np.array): Array of angular velocities.

    Returns:
        Array of shape (..., 4) containing element-wise derivatives of q.

    Example::

        q_prime = rowan.calculus.derivative([1, 0, 0, 0], [1, 0, 0])
    """
    q = np.asarray(q)
    v = np.asarray(v)

    _validate_unit(q)
    return 0.5*multiply(q, _promote_vec(v))


def integrate(q, v, dt):
    R"""Integrate unit quaternions by angular velocity using the following
    equation:

    .. math::
        \dot{q} = \exp\left(\frac{1}{2} \boldsymbol{v} dt\right) q

    Note that this formula uses the `quaternion exponential`_, so the argument
    to the exponential (which appears to be a vector) is promoted to a
    quaternion with scalar part 0 before the exponential is taken.
    A concise derivation is provided in `this paper`_.
    This `webpage`_ contains a more thorough explanation.

    .. _quaternion exponential: https://en.wikipedia.org/wiki/
                                Quaternion#Exponential,_logarithm,_and_power
    .. _this paper: https://www.researchgate.net/publication/
                    260466470_Geometric_Integration_of_Quaternions
    .. _webpage: https://www.ashwinnarayan.com/post/
                 how-to-integrate-quaternions/

    Args:
        q ((...,4) np.array): Array of quaternions.
        v ((...,3) np.array): Array of angular velocities.
        dt ((...) np.array): Array of timesteps.

    Returns:
        Array of shape (..., 4) containing element-wise integrals of q.

    Example::

        v_next = rowan.calculus.integrate([1, 0, 0, 0], [0, 0, 1e-2], 1)
    """
    q = np.asarray(q)
    v = np.asarray(v)
    dt = np.asarray(dt)

    _validate_unit(q)

    return multiply(exp(_promote_vec(v*dt/2)), q)



from setuptools import setup, find_packages

import os

# Gets the version
version = '1.1.7'

# Read README for PyPI, fallback if it fails.
desc = 'Perform quaternion operations using NumPy arrays'
try:
    readme_file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               'README.md')
    with open(readme_file) as f:
        readme = f.read()
except ImportError:
    readme = desc

setup(name='rowan',
      version=version,
      description=desc,
      long_description=readme,
      long_description_content_type='text/markdown',
      url='http://bitbucket.org/glotzer/rowan',
      author='Vyas Ramasubramani',
      author_email='vramasub@umich.edu',
      packages=find_packages(exclude=["tests"]),
      zip_safe=True,
      install_requires=[
          'numpy>=1.10'
      ],
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, <4',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Scientific/Engineering :: Mathematics',
      ],
      )



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
R"""
The rowan package provides a simple interface to slerp, the standard method
of quaternion interpolation for two quaternions.
"""
from __future__ import division, print_function, absolute_import

import numpy as np

from ..functions import power, multiply, conjugate, _validate_unit, log

__all__ = ['slerp',
           'slerp_prime',
           'squad']


def slerp(q0, q1, t, ensure_shortest=True):
    R"""Spherical linear interpolation between p and q.

    The `slerp formula <https://en.wikipedia.org/wiki/Slerp#Quaternion_Slerp>`_
    can be easily expressed in terms of the quaternion exponential (see
    :py:func:`rowan.exp`).

    Args:
        q0 ((...,4) np.array): First array of quaternions.
        q1 ((...,4) np.array): Second array of quaternions.
        t ((...) np.array): Interpolation parameter :math:`\in [0, 1]`
        ensure_shortest (bool): Flip quaternions to ensure we traverse the
            geodesic in the shorter (:math:`<180^{\circ}`) direction.

    .. note::

        Given inputs such that :math:`t\notin [0, 1]`, the values outside the
        range are simply assumed to be 0 or 1 (depending on which side of the
        interval they fall on).

    Returns:
        Array of shape (..., 4) containing the element-wise interpolations
        between p and q.

    Example::

        import numpy as np
        q_slerp = rowan.interpolate.slerp(
            [[1, 0, 0, 0]], [[np.sqrt(2)/2, np.sqrt(2)/2, 0, 0]], 0.5)
    """
    _validate_unit(q0)
    _validate_unit(q1)
    t = np.clip(t, 0, 1)

    q0 = np.asarray(np.atleast_2d(q0))
    q1 = np.array(np.atleast_2d(q1))

    # Ensure that we turn the short way around
    if ensure_shortest:
        cos_theta = np.sum(q0*q1, axis=-1)
        flip = cos_theta < 0
        q1[flip] *= -1

    return multiply(q0, power(multiply(conjugate(q0), q1), t))


def slerp_prime(q0, q1, t, ensure_shortest=True):
    R"""Compute the derivative of slerp.

    Args:
        q0 ((...,4) np.array): First set of quaternions.
        q1 ((...,4) np.array): Second set of quaternions.
        t ((...) np.array): Interpolation parameter :math:`\in [0, 1]`
        ensure_shortest (bool): Flip quaternions to ensure we traverse the
            geodesic in the shorter (:math:`<180^{\circ}`) direction

    Returns:
        An array of shape (..., 4) containing the element-wise derivatives of
        interpolations between p and q.

    Example::

        import numpy as np
        q_slerp_prime rowan.interpolate.slerp_prime(
            [[1, 0, 0, 0]], [[np.sqrt(2)/2, np.sqrt(2)/2, 0, 0]], 0.5)
    """
    _validate_unit(q0)
    _validate_unit(q1)
    t = np.clip(t, 0, 1)

    q0 = np.asarray(np.atleast_2d(q0))
    q1 = np.array(np.atleast_2d(q1))

    # Ensure that we turn the short way around
    if ensure_shortest:
        cos_theta = np.sum(q0*q1, axis=-1)
        flip = cos_theta < 0
        q1[flip] *= -1

    return multiply(
            multiply(q0, power(multiply(conjugate(q0), q1), t)),
            log(multiply(conjugate(q0), q1))
            )


def squad(p, a, b, q, t):
    R"""Cubically interpolate between p and q.

    The SQUAD formula is just a repeated application of Slerp between multiple
    quaternions as originally derived in [Shoemake85]_:

    .. math::
        \begin{equation}
            \textrm{squad}(p, a, b, q, t) = \textrm{slerp}(p, q, t)
            \left(\textrm{slerp}(p, q, t)^{-1}\textrm{slerp}(a, b, t)
            \right)^{2t(1-t)}
        \end{equation}

    .. [Shoemake85] Ken Shoemake. Animating rotation with quaternion curves.
        SIGGRAPH Comput. Graph., 19(3):245-254, July 1985.

    Args:
        p ((...,4) np.array): First endpoint of interpolation.
        a ((...,4) np.array): First control point of interpolation.
        b ((...,4) np.array): Second control point of interpolation.
        q ((...,4) np.array): Second endpoint of interpolation.
        t ((...) np.array): Interpolation parameter :math:`t \in [0, 1]`.

    Returns:
        An array containing the element-wise interpolations between p and q.

    Example::

        import numpy as np
        q_squad = rowan.interpolate.squad(
            [1, 0, 0, 0], [np.sqrt(2)/2, np.sqrt(2)/2, 0, 0],
            [0, np.sqrt(2)/2, np.sqrt(2)/2, 0],
            [0, 0, np.sqrt(2)/2, np.sqrt(2)/2], 0.5)
    """
    _validate_unit(p)
    _validate_unit(a)
    _validate_unit(b)
    _validate_unit(q)
    t = np.clip(t, 0, 1)

    return slerp(slerp(p, q, t, ensure_shortest=False),
                 slerp(a, b, t, ensure_shortest=False),
                 2*t*(1-t),
                 ensure_shortest=False)
